import os
from typing import Iterable, List, Literal, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
import torch
import torch.nn as nn
import torch_dct as dct
from torch.utils.data import TensorDataset, DataLoader


from ..utils.errors import NotFittedError
from ..utils.ot_tools import free_support_bar, self_mapping, self_mapping_single_pass
from ..utils.training import train_autospectro, train_csc
from ..utils.utils import (
    lagged_ts_list,
    flatten_lagged_score,
    multiscale_dataset,
    split_3Darray,
    ssa,
    upsampling_list,
    list_fusion_scales,
    find_consec_ind,
)


class DiLAnoDetectm(torch.nn.Module):
    """Class computing a sparse representation of a dataset in different
    spectral bands at different scales and computes anomaly scores.

    Attributes: 
        nb_scales (int): Number of wavelet scales for decomposing time 
            series.
        wvlt_name (str): Wavelet type to use. Defautl: 'db2'.
        anomaly_detectors_list (List[DilAnoDetect]): List of wrapped
            DilAnoDetect models.
    """

    def __init__(
        self,
        nb_atoms: int,
        nb_blocks: int,
        kernel_size: int,
        layers_sizes: List[int],
        nb_scales: int,
        overlap_perc: float,
        activation: str = 'id',
        wvlt_name: str = 'db2',
        same: bool = False,
        share_weights: bool = False,
        soft: bool = False,
        shrink: bool = True,
        b: float = 1.5,
    ) -> None:
        """Initialize the DiLAnoDetectm object.

        Args: 
            nb_atoms: Number of patterns to learn.
            nb_blocks: Number of unrolled gradient steps blocks.
            kernel_size: Convolutional patterns size, must be odd.
            layers_sizes: Spectral autoencoder layers sizes.
                Layers_sizes[0] defines the sliding window size for
                perfoming DCT.
            nb_scales: Number of wavelet scales for decomposing time 
                series.
            overlap_perc: Overlapping ratio between sliding windows 
                (should be comprised between 0 and 1).
            activation: Activation function at the output of the blocks.
            wvlt_name: Wavelet type to use.
            same: Whether the blocks are identical or have different
                weights.
            share_weights: Whether within a given block, the conv and
                conv_transpose layers share weights or not.
            soft: Whether the shrinking is soft (if not, then it is
                hard).
            shrink: Whether a shrinking activation function is applied
                to each block's output.
            b : DCT sliding window will divided with a factor growing 
                as a power of b.
        """
        super(DiLAnoDetectm, self).__init__()

        self.nb_scales = nb_scales
        self.wvlt_name = wvlt_name
        anomaly_detectors = []
        for i in range(self.nb_scales):
            layers_sizes_i = layers_sizes.copy()
            layers_sizes_i[0] = int(layers_sizes_i[0]/(b**(self.nb_scales-i)))

            anomaly_detectors.append(DiLAnoDetect(
                nb_atoms,
                nb_blocks,
                kernel_size,
                layers_sizes_i,
                overlap_perc,
                activation=activation,
                same=same,
                share_weights=share_weights,
                soft=soft,
                shrink=shrink,
            ))
        self.anomaly_detectors_list = nn.ModuleList(anomaly_detectors)

    def fit(
        self,
        x: List[List[np.array]],
        learning_rate_pauto: float = 0.1,
        order: int = 2,
        epochs_pauto: int = 100,
        chunksize: int = 100,
        learning_rate_csc: float = 1.,
        epochs_csc: int = 5,
        model_name_csc: str = "csc_model_test.pt",
        model_path: str = "models/",
        verbose: bool = False,
    ) -> None:
        """Learn sparse coding models at different scales.

        Args: 
            x: Training signal provided as list of list of 2D numpy
                arrays.
            learning_rate_pauto: Spectral autoencoder learning rate.
            order: Order of the norm to be used when normalizing sparse
                coding model weights.
            epochs_pauto: Number of training epoch when learning
                spectral autoencoder.
            chunksize: Spectral autoencoder training mini batches size.
            learning_rate_csc: Convolutional sparse coding learning
                rate.
            epoch_csc: Number of training epoch when learning
                convolutional sparse coding model.
            model_name_csc: Output convolutional sparse coding model
                file name.
            model_path: Output directory for saving models.
            verbose: Whether to display details about the operations
                done.
        """
        if not os.path.exists(model_path):
            os.makedirs(model_path)

        list_scales = multiscale_dataset(
            x, nb_scales=self.nb_scales, wvlt_name=self.wvlt_name)
        count = 0

        for i, dil in enumerate(self.anomaly_detectors_list):
            if verbose:
                print("Scale ", self.nb_scales-i)
            flag = dil.fit(
                list_scales[i+1],
                learning_rate_pauto=learning_rate_pauto,
                order=order,
                epochs_pauto=epochs_pauto,
                model_name_pauto='auto_model'+str(i)+'.pt',
                chunksize=chunksize,
                learning_rate_csc=learning_rate_csc,
                epochs_csc=epochs_csc,
                model_name_csc='csc_model_test'+str(i)+'.pt',
                model_path=model_path,
            )
            count += flag
        self.nb_scales += count  # Limiting scales coarseness
        for _ in range(int(-count)):
            self.anomaly_detectors_list.pop(0)  # Removing unfitted models

        torch.save(self.state_dict(), model_path+model_name_csc)

    def score_samples(
        self,
        x: List[List[np.array]],
        multivar: bool = False,
        single_pass: bool = True,
    ) -> List[List[np.array]]:
        """Compute anomaly scores.

        Args:
            x: Training signal provided as list of list of 2D numpy
                arrays.
            multivar: Joint monovariate or multivariate anomaly
                detection.
            single_pass: Whether to compute optimal transport mapping
                costs of a dataset onto itself forbidding mass
                immobility.

        Returns:
            The anomaly scores provided as a list of list of 2D numpy
                arrays.
        """
        list_scales = multiscale_dataset(
            x, nb_scales=self.nb_scales, wvlt_name=self.wvlt_name
        )
        list_scores_assets = []
        if x[0][0].shape[0] == 1:
            multivar = False

        for i_asset in range(len(x)):
            list_scores = []
            for i, dil in enumerate(self.anomaly_detectors_list):
                score_i = dil.score_samples(
                    list_scales[i+1][i_asset],
                    multivar=multivar,
                    single_pass=single_pass)
                D = 2**(self.nb_scales-i)
                list_scores.append(upsampling_list(score_i, D))
            list_scores_assets.append(list_fusion_scales(list_scores))

        return list_scores_assets

    def extract_segments(
        self,
        ano_scores: List[List[np.array]],
        thresh: float = 0.8,
    ) -> pd.DataFrame:
        """Extract segments with anomaly scores above a provided
        threshold.

        Args:
            ano_scores: Anomaly scores provided as a list of list of 2D
                arrays.
            thresh: Between 0 and 1. Anomaly detection threshold.

        Returns:
            A dataframe in which each raw is an anomalous segment with
            its different features (length, starting and ending indices,
            average anomaly score, asset, sequence and variable
            indices).
        """
        list_keys = [
            "start", "end", "mean score", "length", "asset", "sequence", "variable"
        ]
        dict_ano = {}
        for key in list_keys:
            dict_ano[key] = []

        for i, ano_asset in enumerate(ano_scores):
            for j, ano_seq in enumerate(ano_asset):
                for k in range(ano_seq.shape[0]):
                    ind = np.where(ano_seq[k, :] >= thresh)[0]
                    if len(ind > 0):
                        ind_seq = find_consec_ind(ind, singletonrm=True)
                        for seq_thresh in ind_seq:
                            list_feat = [
                                seq_thresh[0],
                                seq_thresh[-1] + 1,
                                ano_seq[k, seq_thresh].mean(),
                                len(seq_thresh),
                                i,
                                j,
                                k,
                            ]
                            for ik, key in enumerate(list_keys):
                                dict_ano[key].append(list_feat[ik])

        df = pd.DataFrame(dict_ano)
        df = df.query('length>1')
        return df


class PositiveAutoencoder(torch.nn.Module):
    """Class implementing a simple autoencoder aimed at performing
    dimensionality reduction with positivity constraint on the latent
    representation.

    Attributes:
        layers_sizes (Sequence[int]): Encoder layers sizes. The decoder
            has a single layer of sizes layers_size[-1]xlayers_size[0].
    """

    def __init__(self, layers_sizes: Sequence[int]) -> None:
        """Initialize the PositiveAutoencoder object.

        Args:
            layers_sizes: Encoder layers sizes. the decoder has a single
                layer of sizes layers_size[-1]xlayers_size[0].
        """
        input_size = layers_sizes[0]
        hidden_size = layers_sizes[-1]
        super(PositiveAutoencoder, self).__init__()

        encoder = []

        nb_layers = len(layers_sizes)
        for i in range(nb_layers-1):
            encoder.append(
                nn.Linear(layers_sizes[i], layers_sizes[i+1], bias=True))
            encoder.append(nn.ReLU())
        self.encoder = nn.Sequential(*encoder)
        self.decoder = nn.Linear(hidden_size, input_size, bias=True)

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute a feedforward through the model.

        Args:
            x: Samples batch.

        Returns:
            The input's approximation.
            The input's latent representation.
        """
        compressed = self.encoder(x)
        output = self.decoder(compressed)

        return output, compressed


class PositiveSpectrogramAutoencoder(torch.nn.Module):
    """Class performing a spectral dimension reduction.

    Attributes:
        p_auto (PositiveAutoencoder): Wrapped PositiveAutoencoder, 
            instantiated from a layers_sizes init parameter.
        hidden_size (int): Hidden size, corresponding to the last
            element of the layers_sizes init parameter.
    """

    def __init__(self, layers_sizes: Sequence[int]) -> None:
        """Initialize the PositiveSpectrogramAutoencoder object.

        Args:
            layers_sizes: Encoder layers sizes. the decoder has a single
                layer of sizes layers_size[-1]xlayers_size[0].
        """
        super(PositiveSpectrogramAutoencoder, self).__init__()
        self.p_auto = PositiveAutoencoder(layers_sizes)
        self.hidden_size = layers_sizes[-1]

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute a feedforward through the model.

        Args:
            x: Samples batch.

        Returns:
            The input's approximation.
            compressed: The input's compressed spectral representation.
        """
        dct_x = dct.dct(x, norm='ortho')
        dct_sign = torch.sign(dct_x)
        dct_auto, compressed = self.p_auto(dct_x[:, :, 1:]*dct_sign[:, :, 1:])
        output = dct_auto * dct_sign[:, :, 1:]
        output = torch.cat((dct_x[:, :, 0].unsqueeze(2), output), axis=2)
        output = dct.idct(output, norm='ortho')
        return output, compressed


class DiLAnoDetect(torch.nn.Module):
    """Class computing a sparse representation of a dataset in different
    spectral bands, and anomaly scores.

    Attributes:
        p_auto_spec (PositiveSpectrogramAutoencoder): Wrapped 
            PositiveSpectrogramAutoencoder, instantiated from a
            layers_sizes init parameter.
        relu (ReLU): ReLU activation module.
        dil (DiL): Wrapped DiL model.
        hidden_size (int): Hidden size (number of bands).
        nb_atoms (int): Number of patterns to learn.
        lag (int): Corresponds to layers_sizes[0] + 1.
        overlap_perc (float): Between 0 and 1. Overlapping ratio between
            sliding windows.
        kernel_size (int): Convolutional patterns size, must be odd.
        fitted (bool): Indicates whether the instance has been fitted
            with data.
    """

    def __init__(
        self,
        nb_atoms: int,
        nb_blocks: int,
        kernel_size: int,
        layers_sizes: Sequence[int],
        overlap_perc: float,
        x: Optional[np.array] = None,
        activation: str = 'id',
        same: bool = False,
        share_weights: bool = False,
        soft: bool = False,
        shrink: bool = True,
    ) -> None:
        """Compute a sparse representation of a dataset in different
        spectral bands, model the new representation and compute anomaly
        scores.

        Args: 
            nb_atoms: The number of patterns to learn.
            nb_blocks: The number of unrolled gradient steps blocks. 
            kernel_size: Convolutional patterns size, must be odd.
            layers_sizes (Sequence[int]): Spectral autoencoder layers
                sizes. Layers_sizes[0] defines the sliding window size
                for perfoming DCT.
            overlap_perc: Between 0 and 1. Overlapping ratio between
                sliding windows (should be comprised between 0 and 1).
            x: Sequence of the signal to encode, to estimate nb_atoms
                and initialize the network weights.
            activation: Activation function at the output of the blocks.
            same: Whether the blocks are identical or have different
                weights.
            share_weights: Whether within a given block, the conv and
                conv_transpose layers share weights or not.
            soft: Whether the shrinking is soft (if not, then it is
                hard).
            shrink: Whether a shrinking activation function is applied
                to each block's output.
        """
        super(DiLAnoDetect, self).__init__()
        lag = layers_sizes[0] + 1
        nb_bands = layers_sizes[-1]
        self.p_auto_spec = PositiveSpectrogramAutoencoder(layers_sizes)
        self.relu = nn.ReLU()
        self.dil = DiL(
            nb_atoms,
            nb_blocks,
            kernel_size,
            x=x,
            activation=activation,
            same=same,
            share_weights=share_weights,
            soft=soft,
            shrink=shrink,
        )
        self.hidden_size = nb_bands
        self.nb_atoms = nb_atoms
        self.lag = lag
        self.overlap_perc = overlap_perc
        self.kernel_size = kernel_size
        self.fitted = False

    def fit(
        self,
        x: List[List[np.array]],
        learning_rate_pauto: float = 0.1,
        order: int = 2,
        epochs_pauto: int = 100,
        model_name_pauto: str = "auto_model.pt",
        chunksize: int = 30,
        learning_rate_csc: float = 1.,
        epochs_csc: int = 5,
        model_name_csc: str = "csc_model_test.pt",
        model_path: str = "models/",
        verbose: bool = False,
    ) -> Literal[-1, 0]:
        """Learn sparse coding models.

        Args: 
            x: Training signal provided as list of list of 2D numpy
                arrays.
            learning_rate_pauto: Spectral autoencoder learning rate. 
            order: Order of the norm to be used when normalizing sparse
                coding model weights.
            epochs_auto: Number of training epoch when learning spectral
                autoencoder.
            model_name_pauto: Output spectral autoencoder model file
                name.
            chunksize: Spectral autoencoder training mini batches size.
            learning_rate_csc: Convolutional sparse coding learning
                rate.
            epoch_csc: Number of training epoch when learning
                convolutional sparse coding model.
            model_name_csc: Output convolutional sparse coding model
                file name.
            model_path: Output directory for saving models.
            verbose: Whether to display details about the operations
                done.

        Returns:
            0 whenever fitting is performed, -1 otherwise.
        """
        for param in self.dil.parameters():
            param.requires_grad = True

        for param in self.p_auto_spec.parameters():
            param.requires_grad = True

        hop_length = int(self.lag * (1-self.overlap_perc))

        list_lagged_cubes = []
        for l_asset in x:
            lagged_cube = []
            for seq in l_asset:
                lagged_cube_i = lagged_ts_list(
                    seq, self.lag, hop_length=hop_length, verbose=verbose
                )

                if lagged_cube_i is not None:
                    lagged_cube.append(lagged_cube_i)

                lagged_cube.append(lagged_cube_i)
            if len(lagged_cube) > 0:
                list_lagged_cubes.append(lagged_cube)

        if len(list_lagged_cubes) > 0:
            if verbose:
                print("Learning filters bank")
            b = None
            for l_asset in list_lagged_cubes:
                for seq in l_asset:
                    a = torch.tensor(seq)
                    a = a.transpose(1, 2)
                    if b is None:
                        b = a.detach()
                    else:
                        b = torch.cat((b, a), axis=0)

            b = b.type(torch.float32)
            dataset = TensorDataset(b, b)
            dataloader = DataLoader(dataset, batch_size=100, shuffle=True)

            optimizer_pauto = torch.optim.SGD(
                self.p_auto_spec.parameters(), lr=learning_rate_pauto
            )

            loss_fn = nn.MSELoss()

            for t in range(epochs_pauto):
                if verbose:
                    print(f"Epoch {t+1}\n-------------------------------")
                train_autospectro(
                    dataloader,
                    self.p_auto_spec,
                    loss_fn,
                    optimizer_pauto,
                    path=model_path,
                    model_name=model_name_pauto,
                    order=order,
                )
            if verbose:
                print("Done!")
                print("Learning CSC model")

            mat_chunks = None
            for l_asset in list_lagged_cubes:
                for seq in l_asset:

                    a = torch.tensor(seq)
                    a = a.transpose(1, 2)
                    if mat_chunks is None:
                        mat_chunks = split_3Darray(
                            a, axis=1, chunksize=chunksize)
                    else:
                        mat_chunks = torch.concatenate(
                            (mat_chunks, split_3Darray(
                                a, axis=1, chunksize=chunksize)),
                            axis=1,
                        )
            mat_chunks = mat_chunks.type(torch.float32)
            mat_chunks = mat_chunks.transpose(0, 1)

            for param in self.p_auto_spec.parameters():
                param.requires_grad = False

            dataset = TensorDataset(mat_chunks, mat_chunks)
            dataloader = DataLoader(dataset, batch_size=1, shuffle=True)

            optimizer_csc = torch.optim.SGD(
                self.dil.parameters(), lr=learning_rate_csc
            )
            loss_fn = nn.MSELoss()
            for t in range(epochs_csc):
                if verbose:
                    print(f"Epoch {t+1}\n-------------------------------")
                train_csc(
                    dataloader,
                    self,
                    loss_fn,
                    optimizer_csc,
                    path=model_path,
                    model_name=model_name_csc,
                    positive_weights=True,
                    normalize=True,
                    verbose=verbose,
                )
            if verbose:
                print("Done!")
            self.fitted = True
            return 0
        else:
            return -1

    def score_samples(
        self,
        x: List[List[np.array]],
        multivar: bool = False,
        single_pass: bool = True,
        verbose: bool = False,
    ) -> List[List[np.array]]:
        """Compute anomaly scores.

        Args:
            x: Training signal provided as list of list of 2D numpy
                arrays.
            multivar: Joint monovariate or multivariate anomaly
                detection.
            verbose: Whether to display details about the operations
                done.

        Returns:
            The anomaly scores provided as a list of list of 2D numpy
                arrays.
        """
        if self.fitted is False:
            raise NotFittedError(
                "Please first call the fit method to learn the "
                "convolutional sparse coding model parameters."
            )
        hop_length = int(self.lag*(1-self.overlap_perc))
        list_seq = []

        for seq in x:
            lagged_cube_i = lagged_ts_list(
                seq, self.lag, hop_length=hop_length, verbose=verbose
            )
            list_seq.append(lagged_cube_i)

        if multivar:
            anomaly_scores = self.pointwise_anomaly_score_ot_bar(list_seq)
        else:
            anomaly_scores = self.pointwise_anomaly_score_ot(
                list_seq, single_pass=single_pass
            )

        return anomaly_scores

    def forward(self, xin: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute a feedforward through the model.

        Args:
            xin: Samples batch.

        Returns:
            The input's approximation.
            The input's sparse codes in different spectral bands.
        """
        x = xin.transpose(0, 1)
        dct_x = dct.dct(x, norm='ortho')
        dct_sign = torch.sign(dct_x)
        compressed = self.relu(self.p_auto_spec.p_auto.encoder(
            (dct_x[:, :, 1:]*dct_sign[:, :, 1:])))
        compressed = compressed.transpose(0, 2)
        compressed_est = None

        for i in range(compressed.shape[0]):
            compressed_i, _ = self.dil(
                compressed[i, :, :].unsqueeze(0))
            if i == 0:
                compressed_est = compressed_i
            else:
                compressed_est = torch.cat(
                    (compressed_est, compressed_i), axis=0)

        compressed_est = compressed_est.transpose(0, 2)
        dct_est = self.p_auto_spec.p_auto.decoder(
            compressed_est)*dct_sign[:, :, 1:]
        output = torch.cat((dct_x[:, :, 0].unsqueeze(2), dct_est), axis=2)
        output = dct.idct(output, norm='ortho')
        output = output.transpose(0, 1)

        return output, compressed

    def get_sparse_codes(self, xin: torch.Tensor) -> torch.Tensor:
        """Compute the input sparse codes.

        Args:
            xin: Samples batch.

        Returns:
            The input's sparse codes in different spectral bands.
        """
        sparse_codes = None
        for j in range(xin.shape[0]):
            x = xin[j, :, :].unsqueeze(0)
            x = x.transpose(0, 1)
            dct_x = dct.dct(x, norm='ortho')
            dct_sign = torch.sign(dct_x)
            compressed = self.relu(self.p_auto_spec.p_auto.encoder(
                (dct_x[:, :, 1:]*dct_sign[:, :, 1:])))
            compressed = compressed.transpose(0, 2)
            sparse_codes_j = None
            for i in range(compressed.shape[0]):
                sparse_codes_i = self.dil.get_sparse_codes(
                    compressed[i, :, :].unsqueeze(0)
                )

                if sparse_codes_j is None:
                    sparse_codes_j = sparse_codes_i
                else:
                    sparse_codes_j = torch.cat(
                        (sparse_codes_j, sparse_codes_i), axis=0
                    )
            if sparse_codes is None:
                sparse_codes = sparse_codes_j

            else:
                sparse_codes = torch.cat(
                    (sparse_codes, sparse_codes_j), axis=2
                )
        return sparse_codes

    def pointwise_anomaly_score_ot(
        self,
        list_seq: Iterable[np.array],
        single_pass: bool = False,
    ) -> List[np.array]:
        """Compute joint monovariate anomaly scores using optimal
        transport.

        Args:
            list_seq: Time series provided as an iterable of 2D arrays.
            single_pass: Whether to compute optimal transport mapping
                costs of a dataset onto itself forbidding mass
                immobility.
            Caution: If the first dimension of sparse_codes_freq[i] exceeds a few thousands, 
            the program might run into memory error due to M matrix size in self_mapping 
            or self_mapping_single_pass methods.

        Returns:
            The time series of anomaly scores.
        """
        t_cost_seq = []
        ind_list_seq = []
        ind_list_seq_r = []
        len_seq = []
        self.Exponential_ot = []
        self.OT_cost_interp = []
        sparse_codes_freq = [None for _ in range(self.hidden_size)]
        nb_points = np.zeros((self.hidden_size,))

        for x in list_seq:
            nb_ts = x.shape[0]
            nb_sig = x.shape[-1]
            xin = torch.tensor(x)
            xin = xin.transpose(1, 2)
            xin = xin.transpose(0, 1)
            xin = xin.type(torch.float32)
            sparse_codes = self.get_sparse_codes(xin)
            t_cost_freq = []
            len_freq = []
            ind_list_freq = []
            ind_list_freq_r = []
            
            for i in range(self.hidden_size):
                sc_i = sparse_codes[i, :, :].transpose(-1, 0).detach().numpy()
                h = np.linalg.norm(sc_i, axis=1)
                ind_i_all = np.where(h > 0)[0]
                ind_i_list = []
                ind_i_list_r = []

                len_freq.append(nb_ts)
                for k in range(nb_sig):
                    ind_ik = np.where(h[k*nb_ts:(k+1)*nb_ts] > 0)[0]
                    ind_i_list.append(
                        np.array(np.arange(len(ind_ik)) + nb_points[i]).astype(int))
                    ind_i_list_r.append(ind_ik)
                    nb_points[i] += len(ind_ik)

                ind_list_freq.append(ind_i_list)
                ind_list_freq_r.append(ind_i_list_r)
                if sparse_codes_freq[i] is None:
                    sparse_codes_freq[i] = sc_i[ind_i_all, :]
                else:
                    sparse_codes_freq[i] = np.concatenate(
                        (sparse_codes_freq[i], sc_i[ind_i_all, :]), axis=0
                    )

            ind_list_seq.append(ind_list_freq)
            len_seq.append(len_freq)
            ind_list_seq_r.append(ind_list_freq_r)

        for i in range(self.hidden_size):
            if single_pass:
                t_cost_freq = self_mapping_single_pass(sparse_codes_freq[i])
            else:
                t_cost_freq = self_mapping(sparse_codes_freq[i])
            expo = Exponential()
            expo.fit(np.array(t_cost_freq))
            self.Exponential_ot.append(expo)
            t_cost_seq.append(t_cost_freq)
        
        anomaly_scores_vect = []
        hop_length = int(self.lag*(1-self.overlap_perc))

        for j, ind_select_freq in enumerate(ind_list_seq):
            ano_sig = []
            ano_sig_vect = []
            for k in range(nb_sig):
                ano_sig_i = []
                ano_sig_i_vect = []
                for i in range(self.hidden_size):
                    cost = t_cost_seq[i][ind_select_freq[i][k]]

                    ano = 1-self.Exponential_ot[i].cumul_function(cost)
                    ano_vect = np.zeros((len_seq[j][i]))
                    for count, p in enumerate(ind_list_seq_r[j][i][k]):
                        p_max = int(min(len_seq[j][i], p+self.kernel_size+1))
                        ano_vect[p:p_max] = ano[count]
                    ano_sig_i_vect.append(flatten_lagged_score(
                        ano_vect, self.lag, hop_length
                    ))
                ano_sig.append(ano_sig_i)
                ano_sig_vect.append(np.array(ano_sig_i_vect).max(axis=0))
                print(len(ano_sig_vect))
            anomaly_scores_vect.append(np.array(ano_sig_vect))
        
        return anomaly_scores_vect

    def pointwise_anomaly_score_ot_bar(
        self,
        list_seq: Iterable[np.array],
        verbose: bool = False,
    ) -> List[np.array]:
        """Compute multivariate anomaly scores using Wasserstein
        barycenters.

        Args:
            list_seq: Time series provided as an iterable of 2D arrays.
            verbose: Whether to display details about the operations
                done.
            Caution: If the first dimension of one of the arrays in sparse_codes_freq[i]
            exceeds a few thousands, the program might run into memory error due to M matrix size
            in free_support_bar method.
        Returns:
            The time series of anomaly scores.
        """
        t_cost_seq = []
        ind_list_seq = []
        ind_list_seq_r = []
        len_seq = []
        self.Exponential_ot = []
        self.OT_cost_interp = []

        nb_sig = list_seq[0].shape[-1]
        nb_points = np.zeros((self.hidden_size, nb_sig))
        sparse_codes_freq = [[None for j in range(
            nb_sig)] for i in range(self.hidden_size)]

        for x in list_seq:
            nb_ts = x.shape[0]

            xin = torch.tensor(x)
            xin = xin.transpose(1, 2)
            xin = xin.transpose(0, 1)
            xin = xin.type(torch.float32)
            sparse_codes = self.get_sparse_codes(xin)

            t_cost_freq = []
            len_freq = []
            ind_list_freq = []
            ind_list_freq_r = []

            for i in range(self.hidden_size):
                sc_i = sparse_codes[i, :, :].transpose(-1, 0).detach().numpy()
                h = np.linalg.norm(sc_i, axis=1)
                ind_i_list = []
                ind_i_list_r = []

                len_freq.append(nb_ts)

                for k in range(nb_sig):
                    ind_ik = np.where(h[k*nb_ts:(k+1)*nb_ts] > 0)[0]
                    ind_i_list.append(
                        np.array(np.arange(len(ind_ik))+nb_points[i][k]).astype(int))
                    if sparse_codes_freq[i][k] is None:
                        sparse_codes_freq[i][k] = sc_i[ind_ik+k*nb_ts, :]
                    else:
                        sparse_codes_freq[i][k] = np.concatenate(
                            (sparse_codes_freq[i][k], sc_i[ind_ik+k*nb_ts, :]), axis=0)
                    ind_i_list_r.append(ind_ik)
                    nb_points[i][k] += len(ind_ik)

                ind_list_freq.append(ind_i_list)
                ind_list_freq_r.append(ind_i_list_r)

            ind_list_seq.append(ind_list_freq)
            len_seq.append(len_freq)
            ind_list_seq_r.append(ind_list_freq_r)

        bar_seq = []
        t_cost_seq = []
        for i in range(self.hidden_size):
            l = int(np.median([sparse_codes_freq[i][k].shape[0]
                    for k in range(nb_sig)]))

            sc_i = sparse_codes_freq[i][0]
            for k in range(1, nb_sig):
                sc_i = np.concatenate((sc_i, sparse_codes_freq[i][k]), axis=0)
            if verbose:
                print("bar_shape", sc_i.shape)
            clustering = KMeans(
                n_clusters=l,
                random_state=0,
                n_init="auto"
            ).fit(sc_i)

            list_medoids = []
            labels = list(set(clustering.labels_))
            if -1 in labels:
                labels.remove(-1)
            for l in labels:
                ind_l = np.where(clustering.labels_ == l)[0]
                sparse_codes_l = sc_i[ind_l, :]
                medoid = np.median(sparse_codes_l, axis=0)
                list_medoids.append(medoid)

            medoids = np.array(list_medoids)
            w = np.ones((nb_sig,))/nb_sig

            bar_freq, t_cost_freq = free_support_bar(
                sparse_codes_freq[i], medoids, verbose=verbose
            )
            bar_seq.append(bar_freq)
            t_cost_seq.append(t_cost_freq)

        self.Exponential_bar = []
        for i in range(self.hidden_size):
            t_cost_freq = []
            for t_cost in t_cost_seq[i]:
                t_cost_freq += list(t_cost)
            expo = Exponential()
            expo.fit(np.array(t_cost_freq))
            self.Exponential_bar.append(expo)

        anomaly_scores_vect = []
        hop_length = int(self.lag*(1-self.overlap_perc))

        for j, ind_select_freq in enumerate(ind_list_seq):
            ano_sig = []
            ano_sig_vect = []
            for k in range(nb_sig):
                ano_sig_i = []
                ano_sig_i_vect = []
                for i in range(self.hidden_size):
                    cost = t_cost_seq[i][k][ind_select_freq[i][k]]

                    ano = 1-self.Exponential_bar[i].cumul_function(cost)
                    ano_vect = np.zeros((len_seq[j][i]))
                    for count, p in enumerate(ind_list_seq_r[j][i][k]):
                        p_min = int(max(0, p-(self.kernel_size-1)/2))
                        p_max = int(
                            min(len_seq[j][i], p+(self.kernel_size-1)/2+1)
                        )
                        ano_vect[p_min:p_max] = ano[count]
                    ano_sig_i_vect.append(flatten_lagged_score(
                        ano_vect, self.lag, hop_length
                    ))
                ano_sig.append(ano_sig_i)
                ano_sig_vect.append(np.array(ano_sig_i_vect).max(axis=0))
            anomaly_scores_vect.append(np.array(ano_sig_vect))

        return anomaly_scores_vect


class DiL(nn.Module):
    """Class performing a convolutional sparse coding of a 1D signal.

    Attributes:
        nb_blocks (int): Number of unrolled gradient steps blocks 
        same (bool): Determines whether the blocks are identical or have
            different weights.
        res_layers (List[ResLayer]): List of wrapped ResLayers objects.
    """

    def __init__(
        self,
        nb_atoms: int,
        nb_blocks: int,
        kernel_size: int,
        x: Optional[np.array] = None,
        activation: str = 'id',
        same: bool = False,
        share_weights: bool = False,
        soft: bool = False,
        shrink: bool = True,
    ) -> None:
        """Initialize the DiL object.

        Args:
            nb_atoms: Number of patterns to learn.
            nb_blocks: Number of unrolled gradient steps blocks 
            kernel_size: Convolutional patterns size, must be odd.
            x: Sequence of the signal to encode, to estimate nb_atoms
                and initialize the network weights.
            activation: Activation function at the output of the blocks.
            same: Whether the blocks are identical or have different
                weights.
            share_weights: Whether, within a given block, the conv and
                conv_transpose layers share weights or not.
            soft: Whether the shrinking is soft (if not, then it is
                hard).
            shrink: Whether a shrinking activation function is applied
                to each block's output.  
        """
        super(DiL, self).__init__()

        kernel_size = 2*int(kernel_size/2) + 1
        if x is not None:
            atoms = ssa(x, kernel_size, thresh=5)
            nb_atoms = min(nb_atoms, atoms.shape[0])
        else:
            atoms = None
        self.nb_blocks = nb_blocks
        self.same = same
        if same:
            self.res_layers = ResLayer(
                nb_atoms,
                kernel_size,
                activation=activation,
                x=atoms,
                share_weights=share_weights,
                soft=soft,
                shrink=shrink,
            )
        else:
            self.res_layers = nn.ModuleList()
            for _ in range(nb_blocks):
                self.res_layers.append(ResLayer(
                    nb_atoms,
                    kernel_size,
                    activation=activation,
                    x=atoms,
                    share_weights=share_weights,
                    soft=soft,
                    shrink=shrink,
                ))

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute a feedforward through the model.

        Args:
            x: Samples batch.

        Returns:
            The input's approximation.
            The input's sparse codes.
        """
        if self.same:
            pred, sparse_codes = self.res_layers(x)
        else:
            pred, sparse_codes = self.res_layers[0](x)
        err = x-pred
        for i in range(1, self.nb_blocks):
            if self.same:
                pred, sparse_codes = self.res_layers(err, y=sparse_codes)
            else:
                pred, sparse_codes = self.res_layers[i](err, y=sparse_codes)
            err = x-pred

        return pred, sparse_codes

    def get_sparse_codes(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the input sparse codes.

        Args:
            x: Samples batch.

        Returns:
            The input's sparse codes.
        """
        if self.same:
            pred, sparse_codes = self.res_layers(x)
        else:
            pred, sparse_codes = self.res_layers[0](x)
        err = x-pred
        for i in range(1, self.nb_blocks):
            if self.same:
                pred, sparse_codes = self.res_layers(err, y=sparse_codes)
            else:
                pred, sparse_codes = self.res_layers[i](err, y=sparse_codes)
            err = x-pred

        return sparse_codes


class Exponential:
    """Class estimating the univariate exponential law parameter and
    computing a cumulative distribution function.

    Attributes:
        lambd (Optional[float]): Lambda parameter.
    """

    def __init__(self):
        """Initialize the Exponential object."""
        self.lambd = None

    def fit(self, x: np.array, eps: float = 1.e-32) -> None:
        """Estimate the univariate exponential law parameter.

        Args:
            x: Samples following power law.
            eps: Epsilon parameter.
        """
        self.lambd = 1. / (x.mean() + eps)

    def cumul_function(self, x: np.array) -> np.array:
        """Compute the cumulative distribution function.

        Args:
            x: Samples.

        Returns:
            The cumulative distribution values for the input samples.
        """
        if self.lambd is None:
            raise NotFittedError(
                "Please first call the fit method to determine the "
                "lambd attribute."
            )
        return np.exp(-self.lambd * x)


class ResLayer(nn.Module):
    """Class computing a pseudo gradient step for performing
    convolutional sparse coding using learnt neural network.
    """

    def __init__(
        self,
        nb_atoms: int,
        kernel_size: int,
        x: Optional[torch.Tensor] = None,
        activation: str = "id",
        share_weights: bool = False,
        soft: bool = False,
        shrink: bool = True,
    ) -> None:
        """Initialize a ResLayer object.

        Args: 
            nb_atoms: Number of patterns to learn.
            kernel_size: Convolutional patterns size, must be odd.
            x: Sequence of the signal to encode, to estimate nb_atoms.
            activation: Activation function at the output of the blocks.
            same: Whether the blocks are identical or have different
                weights.
            share_weights: Whether within a given block, the conv and
                conv_transpose layers share weights or not.
            soft: Whether the shrinking is soft (if not, then it is
                hard).
            shrink: Whether a shrinking activation function is applied
                to each block's output.
        """
        super(ResLayer, self).__init__()

        self.softmax = False
        if activation == 'tanh':
            self.activation = nn.Tanh()
        elif activation == 'sigmoid':
            self.activation = nn.Sigmoid()
        elif activation == "softmax":
            self.activation = nn.Softmax(dim=1)
            self.softmax = True
        elif activation == 'id':
            self.activation = nn.Identity()
        self.conv_parallel = ConvParallel(nb_atoms, kernel_size)
        self.conv_parallel_transpose = ConvParallelTranspose(
            nb_atoms, kernel_size)
        self.shrink = shrink
        self.shrink_func = SyncShrink(kernel_size=kernel_size, soft=soft)
        self.relu = ReluParallel()
        if share_weights:
            self.shared_weights = []
            if x is not None:
                if len(x.shape) == 1:
                    atoms = ssa(x, kernel_size, thresh=5)
                else:
                    atoms = np.copy(x)
                nb_atoms = min(nb_atoms, atoms.shape[0])
                for i in range(nb_atoms):
                    self.shared_weights.append(nn.Parameter(
                        (torch.from_numpy(atoms[i, :]).unsqueeze(0)).unsqueeze(0).float()))
            else:
                for i in range(nb_atoms):
                    self.shared_weights.append(
                        nn.Parameter(torch.randn(1, 1, kernel_size)))

            self.conv_parallel = ConvParallel(nb_atoms, kernel_size)
            self.conv_parallel_transpose = ConvParallelTranspose(
                nb_atoms, kernel_size)
            self.conv_parallel.cnn1d[0].weight = self.shared_weights[i]
            self.conv_parallel_transpose.cnn1dT[0].weight = self.shared_weights[i]
            for i, _ in enumerate(self.conv_parallel.cnn1d):
                self.conv_parallel.cnn1d[i].weight = self.shared_weights[i]
                self.conv_parallel_transpose.cnn1dT[i].weight = self.shared_weights[i]

    def forward(
        self, x: torch.Tensor, y: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute a feedforward through the model.

        Args:
            x: Sparse codes batch.
            y: Residuals batch.

        Returns:
            The refined sparse codes.
            The samples estimates.
        """
        x_1 = self.conv_parallel_transpose(x)
        if y is not None:
            x_1 = x_1+y
        if self.softmax:
            x_3 = self.activation(x_1)
        else:
            x_2 = self.relu(x_1)

            x_3 = self.activation(x_2)

        if self.shrink:
            x_3 = self.shrink_func(x_3)

        x_4 = self.conv_parallel(x_3).sum(axis=1)
        x_4 = x_4.unsqueeze(1)

        return x_4, x_3


class ReluParallel(nn.Module):
    """Class computing ReLU channel-wise.

    Attributes:
        relu (ReLU): ReLU activation module.
    """

    def __init__(self) -> None:
        super(ReluParallel, self).__init__()
        self.relu = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute a feedforward through the unit.

        Args:
            x: Input.

        Returns:
            The output after the ReluParallel unit.
        """
        output = torch.cat(
            [
                self.relu(x[:, i, :].unsqueeze(1))
                for i in range(x.shape[1])
            ],
            axis=1,
        )
        return output


class ConvParallel(nn.Module):
    """Parallel convolutional layers.

    Attributes:
        self.cnn1d (ModuleList): ModuleList of 1D CNN modules.
    """

    def __init__(self, nb_atoms: int, kernel_size: int) -> None:
        """Initialize a ConvParallel object.

        Args:
            nb_atoms (int): number of parallel convolutional layers.
            kernel_size (int): filters size.
        """
        super(ConvParallel, self).__init__()
        self.cnn1d = nn.ModuleList()
        for _ in range(nb_atoms):
            cnn1d_i = nn.Conv1d(
                in_channels=1,
                out_channels=1,
                kernel_size=kernel_size,
                stride=1,
                padding="same",
                bias=False,
            )
            self.cnn1d.append(cnn1d_i)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute a feedforward through the model.

        Args:
            x: Samples batch.

        Returns:
            The convolved input.
        """
        if x.shape[1] == 1:
            output = torch.cat(
                [
                    self.cnn1d[i](x)
                    for i in range(len(self.cnn1d))
                ],
                axis=1,
            )
        else:
            output = torch.cat(
                [
                    self.cnn1d[i](x[:, i, :].unsqueeze(1))
                    for i in range(x.shape[1])
                ],
                axis=1,
            )
        return output


class ConvParallelTranspose(nn.Module):
    """Parallel transposed convolutional layers.

    Attributes:
        self.cnn1dT (ModuleList): ModuleList of 1D transposed CNN
            modules.
    """

    def __init__(self, nb_atoms: int, kernel_size: int) -> None:
        """Initialize a ConvParallelTranspose object.

        Args:
            nb_atoms: number of parallel convolutional layers.
            kernel_size: Filter size.
        """
        super(ConvParallelTranspose, self).__init__()
        self.cnn1dT = nn.ModuleList()
        for _ in range(nb_atoms):
            cnn1d_i = nn.ConvTranspose1d(
                in_channels=1,
                out_channels=1,
                kernel_size=kernel_size,
                bias=False,
                padding=int((kernel_size-1) / 2))
            self.cnn1dT.append(cnn1d_i)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute a feedforward through the model.

        Args:
            x: Samples batch.

        Returns:
            The convolved input.
        """
        if x.shape[1] == 1:
            output = torch.cat([self.cnn1dT[i](x)
                               for i in range(len(self.cnn1dT))], axis=1)
        else:
            output = torch.cat([self.cnn1dT[i](x[:, i, :].unsqueeze(1))
                               for i in range(x.shape[1])], axis=1)
        return output


class SyncShrink(nn.Module):
    """Synchronous thresholding layer.

    Attributes:
        kernel_size (int): Thresholding window size.
        soft (bool): Whether the shrinking is soft (if not, then it is
            hard). Default: False.
    """

    def __init__(self, kernel_size: int, soft: bool = False) -> None:
        """Initialize a SyncShrink object.

        Args:
            kernel_size: Thresholding window size.
            soft: Whether the shrinking is soft (if not, then it is
                hard).
        """
        super(SyncShrink, self).__init__()
        self.kernel_size = kernel_size
        self.soft = soft

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Thresholds multichannel input.

        Args:
            x: Multichannel signal.

        Returns:
            The thresholded input.
        """
        padding = int(self.kernel_size/2)
        m = nn.MaxPool1d(2*padding+1, stride=1, padding=padding)
        cumul = torch.abs(x).sum(axis=1)
        output = m(cumul)

        if self.soft:
            delta = (cumul-output).clone().detach()
            w = torch.exp(delta)
        else:
            w = (output-cumul == 0).clone().detach()
            w.to(torch.float64)
        for i in range(x.shape[1]):
            x[:, i, :] = x[:, i, :]*w
        return x
