from typing import List, Sequence, Tuple

import ot
import numpy as np
import random


def ot_mapping(target: np.array, plan: np.array, mode: str = "mean") -> np.array:
    """Compute the barycentric mapping of a point cloud onto another one
    using a transport plan.

    Args:
        target: Target samples array of size nxm, n being the number of
            samples and m the number of features.
        plan: Transport plan of size pxn, p being the number of samples
            to be mapped.
        mode: Mode to be used (mean or other).

    Returns:
        The mapped samples array of size pxm.
    """
    if mode == 'mean':
        nb_samp_max = plan.shape[1]
        X_pushed = plan.dot(target)*nb_samp_max
    else:
        X_pushed = np.zeros((plan.shape[0], target.shape[1]))
        for i in range(plan.shape[0]):
            X_pushed[i, :] = target[plan[i, :].argmax(), :]
    return X_pushed


def free_support_bar(
    list_supports: Sequence[np.array],
    bar: np.array,
    sinkh: bool = False,
    reg: float = 0.001,
    outer_loop: int = 4,
    verbose: bool = False,
) -> Tuple[np.array, List[np.array]]:
    """Compute the wasserstein barycenter of a list of samples sets.

    Caution: If the first dimension of one of the arrays in list_supports or 
        bar exceeds a few thousands, the program might run into memory error 
        due to M matrix size.
     
    Args:
        list_supports: List of samples sets as 2D arrays.
        bar: Initial barycenter.
        sinkh: Whether to use the sinkhorn algorithm or not.
        reg: Entropic regularizer weight in sinkhorn algorithm.
        outer_loop: Number of barycenter update.
        

    Returns:
        The estimated barycenter.
        The list of transportation plans between the barycenter and each
            samples set.
    """
    nb_distributions = len(list_supports)
    nb_points_bar = bar.shape[0]
    d = bar.shape[1]
    transport_cost = []

    for k in range(outer_loop):
        cost = 0
        push = np.zeros((nb_points_bar, d, nb_distributions))
        for i in range(nb_distributions):
            X = list_supports[i]
            M = ot.dist(X, bar)
            nb_samp_i = X.shape[0]
            a, b = np.ones((nb_samp_i,)) / \
                nb_samp_i, np.ones((nb_points_bar,)) / nb_points_bar
            if sinkh:
                G0 = ot.sinkhorn(a, b, M, reg, method='sinkhorn_stabilized')
            else:
                G0 = ot.emd(a, b, M)
            C = M * G0
            cost += C.sum()
            if k == outer_loop-1:
                transport_cost.append(C.sum(axis=1))
            push[:, :, i] = ot_mapping(X, G0.T, mode='mean')
        bar = np.median(push, axis=2)
        if verbose:
            print('Transportation cost: ', cost, ' '+str(k+1)+'/'+str(outer_loop))
            print('Median cost: ', np.median(C), ' Mean cost: ', np.mean(C))

    return bar, transport_cost


def self_mapping(
    x: np.array,
    sinkh: bool = False,
    reg: float = 0.001,
    outer_loop: int = 10,
) -> np.array:
    """Compute average transport mapping costs of a dataset onto itself
    over several random splittings and mappings.

    Caution: If the first dimension of x exceeds a few thousands, 
        the program might run into memory error due to M matrix size.

    Args:
        x: Samples sets as 2D array.
        sinkh: Whether to use the sinkhorn algorithm or not.
        reg: Entropic regularizer weight in sinkhorn algorithm.
        outer_loop: Number of random splittings.
        
    Returns:
        The average transport mapping cost for each sample.
    """
    nb_samp = x.shape[0]
    nb_src = int(nb_samp/2)
    nb_trgt = nb_samp-nb_src
    a, b = np.ones((nb_src,)) / nb_src, np.ones((nb_trgt,)) / nb_trgt
    ind = np.arange(x.shape[0])
    random.shuffle(ind)

    t_cost = np.zeros((nb_samp,))

    for _ in range(outer_loop):
        t_cost_i = np.zeros((nb_samp,))
        src = x[ind[0:int(nb_samp/2)], :]
        trgt = x[ind[int(nb_samp/2):], :]
        M = ot.dist(src, trgt)
        if sinkh:
            G0 = ot.sinkhorn(a, b, M, reg, method="sinkhorn_stabilized")
        else:
            G0 = ot.emd(a, b, M)
        C = M * G0

        t_cost_i[0:nb_src] = C.sum(axis=1)
        t_cost_i[nb_src:] = C.sum(axis=0)

        t_cost[ind] += t_cost_i
        random.shuffle(np.arange(x.shape[0]))

    t_cost /= outer_loop

    return t_cost


def self_mapping_single_pass(x: np.array) -> np.array:
    """Compute optimal transport mapping costs of a dataset onto itself
    forbidding mass immobility.

    Args:
        x: Samples sets as 2D array.
        sinkh: Whether to use the sinkhorn algorithm or not.
        reg: Entropic regularizer weight in sinkhorn algorithm.
        rest_loss: Penalty for keeping mass to oneself.
        Caution: If the first dimension of x exceeds a few thousands, 
        the program might run into memory error due to M matrix size.

    Returns:
        The average transport mapping cost for each sample.
    """
    nb_samp = x.shape[0]
    nb_src = nb_samp
    nb_trgt = nb_samp
    a, b = np.ones((nb_src,)) / nb_src, np.ones((nb_trgt,)) / nb_trgt
    t_cost = np.zeros((nb_samp,))
    src = x
    trgt = x
    M = ot.dist(src, trgt)
    M[np.arange(0, nb_samp), np.arange(0, nb_samp)] = M.max()
    i, j = np.where(M == 0)
    M[i, j] = np.median(M)
    G0 = ot.emd(a, b, M)

    C = M * G0
    t_cost = (C.sum(axis=1)+C.sum(axis=0))/2
    return t_cost
