import torch
import os


def train_autospectro(
    dataloader,
    model,
    loss_fn,
    optimizer,
    path: str = "models",
    model_name: str = "model_default.pt",
    positive_weights: bool = True,
    normalize: bool = True,
    order: int = 2,
    reg: float = 0.1,
    k: int = 1,
) -> None:
    """Train a spectral autoencoder.

    Args:
        dataloader (Torch DataLoader): Training samples mini batches. 
        model (PositiveSpectrogramAutoencoder object): Spectral
            autoencoder model. 
        loss_fn (Torch loss): Training loss.
        optimizer (Torch optimizer): Training optimizer.
        path: model saving directory.
        model_name: output model's name.
        positive_weights: Whether the network's weigths are positive or
            not.
        normalize: Order of the norm to be used when normalizing sparse
            coding model weights.
        reg: Regularizer weight.
        k: Neuron's weights norm's maximum value.
    """
    if not os.path.exists(path):
        os.makedirs(path)
    model.train()
    for batch, (X, y) in enumerate(dataloader):
        # Compute prediction and loss
        pred, _ = model(X)
        m = torch.matmul(list(model.parameters())[0], list(
            model.parameters())[0].transpose(0, 1)).abs()
        diff_m = torch.diff(list(model.parameters())[0], dim=1)
        loss_fit = loss_fn(pred, y)
        loss = loss_fit+reg * \
            (torch.norm(m-torch.eye(m.shape[0]), p=2) +
             (torch.norm(diff_m, dim=1)**2).sum())
        if batch % 100 == 0:
            print(loss_fit, batch)
        # Backpropagation
        loss.backward()
        optimizer.step()
        if positive_weights:
            for p in model.parameters():
                p.data.clamp_(0)
        if normalize:
            for p in model.parameters():
                if torch.norm(p.data) > k:
                    p.data = k*p.data/torch.norm(p.data, p=order)
        optimizer.zero_grad()
        torch.save(model.state_dict(), os.path.join(path, model_name))


def train_csc(
        dataloader,
        model,
        loss_fn,
        optimizer,
        path: str = "models/",
        model_name: str = "model_default.pt",
        positive_weights: bool = True,
        normalize: bool = True,
        k: float = 1.,
        order: int = 2,
        verbose: bool = False,
) -> None:
    """Train a convolutional sparse coding model.

    Args:
        dataloader (Torch DataLoader): Training samples mini batches. 
        model (PositiveSpectrogramAutoencoder object): Spectral
            autoencoder model. 
        loss_fn (Torch loss): Training loss.
        optimizer (Torch optimizer): Training optimizer.
        path: model saving directory.
        model_name: output model's name.
        positive_weights: Whether the network's weigths are positive or
            not.
        normalize: Order of the norm to be used when normalizing sparse
            coding model weights.
        reg: Regularizer weight.
        k: Neuron's weights norm's maximum value.
        verbose: Whether to display details about the operations done.
    """
    if not os.path.exists(path):
        os.makedirs(path)
    model.dil.train()
    for batch, (X, y) in enumerate(dataloader):
        # Compute prediction and loss
        pred, _ = model(X)

        loss = loss_fn(pred, y)
        # Backpropagation
        loss.backward()
        optimizer.step()
        if positive_weights:
            for p in model.dil.parameters():
                p.data.clamp_(0)
        if normalize:
            for p in model.dil.parameters():
                if torch.norm(p.data) > k:
                    p.data = k*p.data/torch.norm(p.data, p=order)

        optimizer.zero_grad()
        if verbose:
            if batch % 1000 == 0:
                print("===== CSC loss: ", loss)

        torch.save(model.state_dict(), os.path.join(path, model_name))
