from typing import Any, Iterable, List, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import torch
import pywt
from numpy.linalg import svd


def lagged_ts(
    ts: np.array,
    lag: int,
    hop_length: int = 1,
    min_nbsegments: int = 50,
    min_lag: int = 64,
    verbose: bool = False,
) -> np.array:
    """Transform a 1D array into a 2D array of overlapping segments from
        the input array.

    Args: 
        ts: Input signal.
        lag: Sliding window size. Should be smaller than ts 
            size/min_nbsegments.
        hop_length: Sliding window jumps length. Should be smaller
            than lag.
        min_nbsegments: Minimal number of segments extracted.
        min_lag: Minimal lag size.
        verbose: Whether to display details about the operations
            done.

    Returns:
        A 2D array of overlapping segments from the ts.
    """
    lag = int(lag)
    hop_length = int(hop_length)
    min_nbsegments = int(min_nbsegments)
    dim = ts.shape[0]

    if lag > dim/min_nbsegments:
        raise ValueError(
            "lag should be less than or equal to min_nbsegments."
        )
    if lag < min_lag:
        raise ValueError(
            "lag should be greater than or equal to min_lag."
        )

    if hop_length > lag:
        hop_length = lag - 1
        if verbose:
            print("hop_length exceed")

    res = np.zeros((int((dim - lag)/hop_length), lag))
    window = np.hanning(lag)
    for i in range(int((dim - lag)/hop_length)):
        res[i, :] = ts[i*hop_length:i*hop_length+lag] * window
    return res


def lagged_ts_list(
    mat: np.array,
    lag: int,
    hop_length: int = 1,
    min_nbsegments: int = 50,
    min_lag: int = 64,
    verbose: bool = False,
) -> np.array:
    """Transform 1D arrays into 2D arrays of overlapping segments from
    the input arrays.

    Args: 
        mat: Input time series arranged as a nxm matrix where n is
            the number of time series.
        lag: Sliding window size. Should be smaller than ts
            size/min_nbsegments.
        hop_length: Sliding window jumps length. Should be smaller
            than lag.
        min_nbsegments: Minimal number of segments extracted.
        min_lag: Minimal lag size.
        verbose: Whether to display details about the operations
            done.

    Returns:
        A 3D array of overlapping segments from the time series.
    """
    lag = int(lag)
    hop_length = int(hop_length)
    min_nbsegments = int(min_nbsegments)
    dim = mat.shape[1]
    if lag > dim/min_nbsegments:
        raise ValueError(
            "lag should be less than or equal to min_nbsegments.")
    if lag < min_lag:
        raise ValueError(
            "lag should be greater than or equal to min_lag."
        )

    if hop_length > lag:
        hop_length = lag - 1
        if verbose:
            print("hop_length exceed")

    output = None
    for i in range(mat.shape[0]):
        res = lagged_ts(mat[i, :], lag, hop_length=hop_length)
        if output is None:
            output = np.expand_dims(res, axis=2)
        else:
            output = np.concatenate(
                (output, np.expand_dims(res, axis=2)), axis=2)
    return output


def flatten_lagged_ts(m: np.array, hop_length: int) -> np.array:
    """Compute 1D time series from a 2D matrix of overlapping segments
    from this time series.

    Args:
        m: 2D matrix of overlapping segments of an unknown time
            series
        hop_length: sliding window jumps length.

    Returns:
        The 1D "flattened" time series.
    """
    nb_hops = m.shape[0]
    lag = m.shape[1]
    size_flat = lag + (nb_hops-1)*hop_length
    flat_signal = np.zeros((size_flat,))
    for i in range(nb_hops):
        flat_signal[i*hop_length:i*hop_length+lag] = m[i, :]
    return flat_signal


def flatten_lagged_score(score: np.array, lag: int, hop_length: int) -> np.array:
    """Compute a flattened time series of anomaly scores for anomaly
    scores computed on overlapping segments of a original time series.

    Args:
        score: Anomaly scores of overlapping segments of a given
            time series
        lag: Sliding window size. Should be smaller than ts
            size/min_nbsegments.
        hop_length: Sliding window jumps length. Should be smaller
            than lag.

    Returns:
        The 1D "flattened" anomaly scores.
    """
    nb_hops = score.shape[0]
    size_flat = lag + (nb_hops-1)*hop_length
    flat_score = np.zeros((size_flat,))
    overlap = lag-hop_length
    for i in range(nb_hops):

        if i > 0:
            flat_score[i*hop_length:i*hop_length+overlap] += 0.5 * score[i]
        flat_score[i*hop_length+overlap:i*hop_length+lag] = score[i]
        if i < nb_hops-1:
            flat_score[i*hop_length+lag-overlap:i*hop_length+lag] *= 0.5

    return flat_score


def multiscale_dataset(
    data_list: Iterable[Sequence[np.array]],
    nb_scales: int = 3,
    wvlt_name: str = "db2",
) -> List[Any]:
    """Compute a wavelet transform of a list of time series.

       Args:
           data_list: Iterable of sequences of 2D arrays of time
               series.
           nb_scales: Number of wavelets scales to compute.
           wvlt_name: Wavelet transform type.

       Returns:
           The list of wavelet decompositions of the input time series.
    """
    wvl = pywt.Wavelet(wvlt_name)
    list_scales = [[] for _ in range(nb_scales+1)]
    for l_asset in data_list:
        list_seq = [[] for _ in range(nb_scales+1)]
        for seq in l_asset:
            C = pywt.wavedec(seq, wvl, level=nb_scales, mode="symmetric")
            for i in range(0, nb_scales+1):
                list_seq[i].append(C[i])
        for i in range(0, nb_scales+1):
            list_scales[i].append(list_seq[i])

    return list_scales


def split_3Darray(cube: torch.Tensor, chunksize: int = 500, axis: int = -1) -> torch.Tensor:
    """Split a 3D tensor along the first dimension and concatenate
    chunks.

    Args:
        cube: 3D data tensor.
        chunksize: Chunk size.
        axis: Concatenation direction.

    Returns:
        The 3D data tensor whose first dimension is equal to the
            block size.
    """
    if chunksize > cube.shape[0]:
        chunksize = cube.shape[0]
        print("Warning: chunksize exceeds first axis size, setting to ",
              cube.shape[0])
    l = cube.shape[0]
    nb_chunks = int(l/chunksize)+1
    output = None
    for i in range(nb_chunks-1):
        if i == 0:
            output = cube[0:chunksize, :, :]
        else:
            output = torch.cat(
                (output, cube[i*chunksize:(i+1)*chunksize, :, :]), axis=axis)
    output = torch.cat((output, cube[-chunksize-1:-1, :, :]), axis=axis)
    return output


def ssa(ts: np.array, size: int, thresh: float = 5., verbose=False) -> np.array:
    """Perform a singular sprectrum analysis.

    Args: 
        ts: Input array.
        size: size of sliding window.
        thresh: percentage of variability to retain in the singular values.
        verbose: Whether to display details about the operations
            done.

    Returns:
        The matrix of selected singular vectors.         
    """
    lag_mat = lagged_ts(ts, size, verbose=verbose)
    _, s, vh = svd(lag_mat.T.dot(lag_mat))
    err_perc = 100 * (1 - np.cumsum(s**2)/np.sum(s**2))
    i = np.where(err_perc < thresh)[0][0]
    output = vh[0:i+1, :]
    return output


def upsampling_mat(t: np.array, D: int) -> np.array:
    """Upsample a 2D array repeating values for padding.

    Args: 
        t: 2D array to upsample along the second dimension.
        D: Upsampling factor.

    Returns:
        t_up: 2D upsampled array. 
    """
    t_up = np.zeros((t.shape[0], t.shape[1]*D))
    ind_range = np.arange((t.shape[1]))*D
    for k in range(D):
        t_up[:, ind_range+k] = t
    return t_up


def upsampling_list(t_list: Iterable[np.array], D: int) -> List[np.array]:
    """Upsample 2D arrays repeating values for padding.

    Args: 
        t: Sequence of 2D arrays to upsample along the second dimension.
        D: Upsampling factor.

    Returns:
        The list of 2D upsampled arrays.
    """
    t_up = [upsampling_mat(t, D) for t in t_list]

    return t_up


def list_fusion_mat(list_mat: Iterable[np.array], method: str = "avg") -> np.array:
    """Perform time step wise max or average pooling over variable
    lengths time series.

    Args: 
        list_mat: Iterable of 2D arrays of potentially different second
            dimensions.
        method: Pooling method. Default: 'avg', for averaging.
            Otherwise, the maximum is taken.

    Returns:
        The 2D array of pooled values. 
    """
    l = np.min([mat.shape[1] for mat in list_mat])
    if method == "avg":
        mat_fus = np.array([mat[:, :l] for mat in list_mat]).mean(axis=0)
    else:
        mat_fus = np.array([mat[:, :l] for mat in list_mat]).max(axis=0)
    return mat_fus


def list_fusion_scales(
    list_scales: Sequence[Sequence[np.array]],
    method: str = "avg",
) -> np.array:
    """Perform time step wise max or average pooling over a list of sets
        of variable lengths time series.

    Args:
        list_scales: Sequence of sequences of 2D arrays of potentially
            different second dimensions
        method: Pooling method. Default: 'avg', for averaging.
            Otherwise, the maximum is taken.

    Returns:
        The list of 2D array of pooled values. 
    """
    list_out = []
    nb_seq = len(list_scales[0])

    for i in range(nb_seq):
        l_mat = [l[i] for l in list_scales]
        list_out.append(list_fusion_mat(l_mat, method=method))

    return list_out


def find_consec_ind(
    ind: Union[np.array, Sequence],
    singletonrm: bool = False,
) -> List[List[int]]:
    """Find sequences of consecutive integers in a 1D array of integers.

    Args: 
        ind: Array or list.
        singletonrm: Whether 1-sized sequences are removed.

    Returns:
            The list of sequences of consecutive integers. 
    """
    seq = [[ind[0]]]
    k = 0
    while k < len(ind)-1:
        if ind[k+1] == ind[k]+1:
            seq[-1].append(ind[k]+1)
        else:
            if singletonrm:
                if len(seq[-1]) == 1:
                    seq.pop()
            seq.append([ind[k+1]])
        k += 1
    return seq


def nans_list(df: pd.DataFrame, thresh: float) -> Tuple[List[List[int]], List[int]]:
    """Find sequences of consecutive indices of corrupted values in a
    dataframe.

    Args: 
        df: Multivariate time series in a dataframe.
        thresh: Threshold above which a value is considered corrupted.

    Returns:
        The list of sequences of consecutive indices of nans.
        The list of number of corrupted variables for each extracted
            index.
    """
    df_remove = (df != df) | (df.abs() > thresh)
    nb_corrupt_var = df_remove.values.astype(int).sum(axis=1)
    ind_nans = np.where(nb_corrupt_var > 1)[0]
    sequences_nans = find_consec_ind(ind_nans)
    nb_var = [nb_corrupt_var[s] for s in sequences_nans]
    return sequences_nans, nb_var


def sample_wise_imput(df: pd.DataFrame, var_thresh: float, value_thresh: float) -> pd.DataFrame:
    """Impute missing values.

    Args: 
        df: Multivariate time series in a dataframe.
        var_thresh: Number of corrupted variable under which the missing
            values are imputed based valid values from other variables
            at the same time.
        value_thresh: Threshold above which a value is considered
            corrupted.

    Returns:
        The dataframe with imputed values.
    """
    sequences_nans, nb_var = nans_list(df, value_thresh)
    for count, s in enumerate(sequences_nans):
        k = np.where(nb_var[count] < var_thresh)[0]
        if len(k) > 0:
            for ind in np.array(s)[k]:
                imput_mask = (df.values[ind, :] != df.values[ind, :]) |\
                    (df.values[ind, :] > value_thresh)
                ref_ind = np.where(imput_mask == 0)[0]
                target_ind = np.where(imput_mask == 1)[0]
                df.values[ind, target_ind] = \
                    np.median(df.values[ind, ref_ind])
    return df


def multiscale_imput(signal: np.array, value_thresh: float) -> Tuple[np.array, List[List[int]]]:
    """Impute missing value based on adaptive smoothing.

    Args: 
        signal: Time series to imput.
        value_thresh: Threshold above which a value is considered
            corrupted.

    Returns:
        The time series with imputed values.
        The list of sequences of consecutive indices of corrupted
            values.
    """
    s_rm = (signal != signal) | (signal > value_thresh)
    ind_nan = np.where(s_rm == True)[0]
    seq = find_consec_ind(ind_nan)
    signal_imput = np.copy(signal)

    l_max = 0
    len_sig = len(signal)
    for s in seq:
        l = len(s)
        imin = max(s[0]-l, 0)
        imin = min(imin, len_sig-3*l)
        imax = imin+3*l
        sig_s = signal[imin:imax]
        ind_s = np.where(sig_s == sig_s)[0]
        signal_imput[s] = np.median(sig_s[ind_s])
        if l > l_max:
            l_max = l
    nb_scales = int(np.log2(l_max))
    sig_wvlt = np.zeros((2**(int(np.log2(len_sig))+1),))
    sig_wvlt[0:len_sig] = signal_imput
    coeffs = pywt.swt(
        sig_wvlt,
        "db2",
        level=nb_scales,
        norm=True,
        trim_approx=False,
    )
    for c in coeffs[1:]:
        for s in seq:
            c[1][s] = 0

    signal_imput_smooth = pywt.iswt(coeffs, "db2", norm=True)
    signal_imput_final = np.copy(signal_imput)
    for s in seq:
        signal_imput_final[s] = signal_imput_smooth[s]

    return signal_imput, seq
